//---------------------------------------------------------------------------------------
/// \file	CircleSelectTool.cpp
/// \author	Rob Bateman
/// \brief	Right, well having a rectangular selection tool is pretty useless. This example
/// 		will add in a custom selection tool that will actually select things in a circle,
/// 		rather than a rectangle.
///
/// 			This process involves a little bit of a hack, basically we :
///
///				* select the items using a rectangular area of the screen.
/// 			* Then query the current selection list via MGlobal
///				* iterate through each component
///					* transform it into 2D coords
/// 				* add any components whose 2D coord falls within the circle to a selection list
/// 			* select the new selection list
///
/// 			Unfortunately, there are a lot of component types that we really need to
/// 		iterate over. To start with then, this example will only iterate over mesh
/// 		vertices and nothing else.
///
/// 			The Tool3 example is a somewhat improved version of the lassoTool that comes
/// 		in the maya API. The only difference is that mine works properly and supports
/// 		all of the component types.
///
/// 			C++ templates might be worth looking at for the component iterating, it
/// 		is likely so save you some time in the long run ;)
/// 		custom context.
///
//---------------------------------------------------------------------------------------

#include "CircleSelectTool.h"
#include <math.h>

//---------------------------------------------------------------------------------------	CircleSelectToolCommand :: makeObj()
/// \brief	This just needs to return a new instance of our tools context. This is used
/// 		by maya to set our tool as current.
/// \return	a new instance of the tool
///
MPxContext* CircleSelectToolCommand::makeObj()
{
	return new CircleSelectTool;
}

//---------------------------------------------------------------------------------------	CircleSelectToolCommand :: creator()
/// \brief	This is used by maya to create a new instance of the tool command
/// \return	a new RectangleSelectContextCmd instance
///
void* CircleSelectToolCommand::creator()
{
	return new CircleSelectToolCommand;
}

/// some help text to display in the help line UI element
const char* g_HelpString = "Click and Drag a circular region";

//---------------------------------------------------------------------------------------	CircleSelectTool :: CircleSelectTool()
/// \brief	ctor
///
CircleSelectTool::CircleSelectTool()
{
	setTitleString ( "Circle Selection Tool" );
	setImage("CircleSelectTool.xpm", MPxContext::kImage1 );
}

//---------------------------------------------------------------------------------------	CircleSelectTool :: ~CircleSelectTool()
/// \brief	dtor
///
CircleSelectTool::~CircleSelectTool()
{
}

//---------------------------------------------------------------------------------------	CircleSelectTool :: toolOnSetup()
/// \brief	This is used to setup any user interface things you need when the tool
/// 		is made current.
/// \param	event	-	information on the input event
///
void CircleSelectTool::toolOnSetup ( MEvent & )
{
	setHelpString( g_HelpString );	//	sets the help text in the help UI item
}

//---------------------------------------------------------------------------------------	CircleSelectTool :: doPress()
/// \brief	overide to handle a mouse press event
/// \param	event	-	information on the input event
/// \return	MS::kSuccess or MS::kFailure
///
MStatus CircleSelectTool::doPress( MEvent & event )
{
	// if we have a left mouse click, start the selection of
	// our rectangular region
	//
	if(event.mouseButton() == MEvent::kLeftMouse)
	{
		// We need to keep a reference of the view in which this command
		// was called. This basically lets us know which viewport to draw
		// in...
		m_View = M3dView::active3dView();

		// we'll only handle the event if the left mouse button is the only thing
		// held down.
		if( !event.isModifierShift() &&
			!event.isModifierControl() &&
			!event.isModifierMiddleMouseButton() )
		{
			// get the current position of the cursor. We need to know this so we can
			// start drawing our selection rectangle!
			event.getPosition( m_Start[0], m_Start[1] );

			// make sure both the start and end points are in the same place.
			m_End[0] = m_Start[0];
			m_End[1] = m_Start[1];

			m_Radius=0;

			// yay!
			return MS::kSuccess;
		}
	}

	// just let the base class handle the event
	return MPxContext::doPress(event);
}

//---------------------------------------------------------------------------------------	CircleSelectTool :: doDrag()
/// \brief	overide to handle a mouse drag event
/// \param	event	-	information on the input event
/// \return	MS::kSuccess or MS::kFailure
///
MStatus CircleSelectTool::doDrag( MEvent & event )
{
	// if we are dragging and left mouse button is pressed...
	if(!event.isModifierLeftMouseButton())
	{
		// get the new location of the cursor
		event.getPosition( m_End[0], m_End[1] );

		// calculate the vector between the start and end points
		short dx = m_End[0] - m_Start[0];
		short dy = m_End[1] - m_Start[1];

		float temp = static_cast<float>(dx*dx + dy*dy);
		// a bit of pythagorus should calculate the radius
		m_Radius = static_cast<short>( sqrt( temp ) );

		glClear(GL_COLOR_BUFFER_BIT);

		// to draw the selection circle within maya, we need to
		// draw a custom overlay in the viewport in which the tool
		// is being used.
		//
		// This means we have to surround the
		m_View.beginOverlayDrawing();

			// clear the overlay planes
			m_View.clearOverlayPlane();

			// Set up the orthographic projection matix
			glMatrixMode( GL_PROJECTION );
			glLoadIdentity();
			gluOrtho2D( 0.0f, m_View.portWidth(),
						0.0f, m_View.portHeight() );

			glMatrixMode( GL_MODELVIEW );
			glLoadIdentity();

			// honestly don't know why we need to translate by this magic number
			// amount? oh well, it works....
			//
			glTranslatef(0.375f+m_Start[0], 0.375f+m_Start[1], 0.0f);

			// Set line width
			glLineWidth( 1.0 );

			// Draw selection circle
			glBegin( GL_LINE_LOOP );
				for(int i=0;i<36;++i)
				{
					float x = m_Radius * cos( i*10 * 3.14159f / 180.0f );
					float y = m_Radius * sin( i*10 * 3.14159f / 180.0f );
					glVertex2f(x,y);
				}
			glEnd();

		// done drawing our overlay on the viewport
		m_View.endOverlayDrawing();
	}

	return MPxContext::doDrag(event);
}

/// \brief	 an internal utility function to determine if the specified 2D point is
///		 	within the selection circle
/// \param	x	-	the x coord of the point to set
/// \param	y	-	the y coord of the point to set
/// \return	true if OK
///
bool CircleSelectTool::ShouldAddPoint(const MPoint& point) {
	MStatus stat;
	short x,y;
	m_View.worldToView( point, x, y, &stat );

	if (!stat) {
		stat.perror("Could not get position");
		return false;
	}

	short dx = x - m_Start[0];
	short dy = y - m_Start[1];
	float temp = static_cast<float>( dx*dx + dy*dy );
	short dist = static_cast<short>( sqrt(temp) );

	if ( dist < m_Radius ) {
		return true;
	}
	return false;
}

//---------------------------------------------------------------------------------------	CircleSelectTool :: doRelease()
/// \brief	overide to handle a mouse release event
/// \param	event	-	information on the input event
/// \return	MS::kSuccess or MS::kFailure
///
MStatus CircleSelectTool::doRelease( MEvent & event )
{
	MStatus stat;

	// only bother handling the release of a left mouse button.
	if(event.mouseButton() != MEvent::kLeftMouse) {
		return MPxContext::doRelease(event);
	}

	// Clear the overlay plane & restore from overlay drawing
	m_View.beginOverlayDrawing();
		m_View.clearOverlayPlane();
	m_View.endOverlayDrawing();

	short minx = m_Start[0] - m_Radius;
	short maxx = m_Start[0] + m_Radius;
	short miny = m_Start[1] - m_Radius;
	short maxy = m_Start[1] + m_Radius;

	// Use the selection mechanism within MGlobal to select the items within the
	//
	MGlobal::selectFromScreen(  minx, miny, maxx, maxy, MGlobal::kReplaceList );

	// now then, currently we can query the selection list from MGlobal and
	// it may contain a list of selected objects, or a list of selected components,
	// ie a list of mesh vertices.
	//
	//	We should really query this list and see what is contained within it.
	//

	MSelectionList rectangleList, newList;

	// Get the list of selected items from within the bounding box
	// and create a iterator for them.
	MGlobal::getActiveSelectionList( rectangleList );


	// Iterate over the objects within the bounding box, extract the
	// ones that are within the lasso, and add those to newList.
	MItSelectionList it( rectangleList );
	newList.clear();

	while ( !it.isDone() ) {
		MDagPath	dagPath;
		MObject		component;
		MPoint		point;

		it.getDagPath( dagPath, component );

		// not a component, i'll assume it's an object then....
		if (!component.isNull()) {


			switch (component.apiType()) {



			case MFn::kMeshVertComponent: {

					MItMeshVertex itVertex( dagPath, component, &stat );
					while ( !itVertex.isDone() )
					{
						point = itVertex.position(MSpace::kWorld );

						if ( ShouldAddPoint( point ) )
							newList.add(dagPath, itVertex.vertex());

						itVertex.next();
					}
					break;
				}

			case MFn::kCurveCVComponent: {

					MItCurveCV itCurve( dagPath, component, &stat );
					while ( !itCurve.isDone() )
					{
						point = itCurve.position(MSpace::kWorld );

						if ( ShouldAddPoint( point ) )
							newList.add(dagPath, itCurve.cv());

						itCurve.next();
					}
					break;
				}

			case MFn::kMeshEdgeComponent: {

					MItMeshEdge itEdge( dagPath, component, &stat );
					while ( !itEdge.isDone() )
					{
						point = itEdge.center(MSpace::kWorld );

						if ( ShouldAddPoint( point ) )
							newList.add(dagPath, itEdge.edge());

						itEdge.next();
					}
					break;
				}

			case MFn::kMeshPolygonComponent: {

					MItMeshPolygon itPoly( dagPath, component, &stat );
					while ( !itPoly.isDone() )
					{
						point = itPoly.center(MSpace::kWorld );

						if ( ShouldAddPoint( point ) )
							newList.add(dagPath, itPoly.polygon());

						itPoly.next();
					}

					break;
				}

			case MFn::kSurfaceCVComponent: {

					MItSurfaceCV itCV( dagPath, component, true, &stat );
					while ( !itCV.isDone(NULL) )
					{
						point = itCV.position(MSpace::kWorld );

						if ( ShouldAddPoint( point ) )
							newList.add(dagPath, itCV.cv());

						itCV.next();
					}

					break;
				}

			case MFn::kMeshVtxFaceComponent:
			case MFn::kSubdivFaceComponent:
			case MFn::kSubdivEdgeComponent:
			case MFn::kCurveEPComponent:
			case MFn::kCurveKnotComponent:
			case MFn::kCurveParamComponent:
			case MFn::kIsoparmComponent:
			case MFn::kPivotComponent:
			case MFn::kSurfaceEPComponent:
			case MFn::kSurfaceKnotComponent:
			case MFn::kEdgeComponent:
			case MFn::kSurfaceRangeComponent:
			case MFn::kDecayRegionCapComponent:
			case MFn::kDecayRegionComponent:
			case MFn::kMeshComponent:
			case MFn::kMeshFrEdgeComponent:
			case MFn::kOrientationComponent:
			case MFn::kSubVertexComponent:
			case MFn::kMultiSubVertexComponent:
			case MFn::kSetGroupComponent:
			case MFn::kDynParticleSetComponent:
			case MFn::kSurfaceFaceComponent:
			case MFn::kSubdivMapComponent:
			case MFn::kMeshMapComponent:
			case MFn::kSubdivCVComponent:
			case MFn::kLatticeComponent:
				{
					MItGeometry itGeom( dagPath, component, &stat );

					if(stat != MS::kSuccess)
					{
						std::cerr << "feck" << std::endl;
						break;
					}

					while ( !itGeom.isDone() )
					{
						point = itGeom.position(MSpace::kWorld );

						if ( ShouldAddPoint( point ) )
							newList.add(dagPath, itGeom.component());

						itGeom.next();
					}
					break;
				}

			default:
				break;
			}
		}

		it.next();

	}

	MGlobal::setActiveSelectionList( newList );

	// nullify selection area
	m_Radius = m_Start[0] = m_Start[1] = m_End[0] = m_End[1] = 0;

	//
	return MS::kSuccess;
}

